/*  2012 Matthias Linder (http://matthiaslinder.com) */
/* - Use as you feel fit, but leave this copyright. - */

using System;
using System.IO;
using System.IO.Ports;
using System.Text;
using System.Threading;
using Hardware.GPS.Extensions;
using Hardware.GPS.Helpers;
using Microsoft.SPOT;

namespace Hardware.GPS
{
    /// <summary>
    /// UART implementation of the CN-06 GPS (Neo-6 / U-Blox) Module
    /// </summary>
    /// <remarks>Questions? -> http://matthiaslinder.com</remarks>
    public class CN06GPS
    {
        /// <summary>
        /// Power (Saving) Mode
        /// </summary>
        public enum PowerSetting : byte
        {
            /// <summary>
            /// Maximum Performance (default)
            /// </summary>
            MaxPerformance = 0,

            /// <summary>
            /// Power Savings
            /// </summary>
            PowerSave = 1,

            /// <summary>
            /// Eco Mode
            /// </summary>
            EcoMode = 4,
        }

        /// <summary>
        /// Dynamic Model
        /// </summary>
        public enum DynNavModel : byte
        {
            /// <summary>
            /// Portable (default)
            /// </summary>
            Portable = 0,

            /// <summary>
            /// Stationary item
            /// </summary>
            Stationary = 2,

            /// <summary>
            /// Slowly moving pedestrian
            /// </summary>
            Pedestrian = 3,

            /// <summary>
            /// Automobile
            /// </summary>
            Automotive = 4,

            /// <summary>
            /// Sea
            /// </summary>
            Sea = 5,

            /// <summary>
            /// Airborne with &lt; 1G
            /// </summary>
            Airborne1G = 6,

            /// <summary>
            /// Airborne with &lt; 2G
            /// </summary>
            Airborne2G = 7,

            /// <summary>
            /// Airborne with &lt; 4G
            /// </summary>
            Airborne4G = 8
        }


        /// <summary>
        /// Constant which converts knots to km/h when multiplied
        /// </summary>
        private const double KNOTS_TO_KMH = 1.852;

        private readonly SerialPort _port;
        private readonly char[] _buf = new char[128];
        private int _read;
        private ushort _rate = 500;
        private DynNavModel _navModel = DynNavModel.Portable;
        private PowerSetting _powerMode = PowerSetting.MaxPerformance;

        /// <summary>
        /// Latitude (Y), positive = N, negative = S, WGS84, or 0.0 if unknown
        /// </summary>
        public double Latitude { get; private set; }

        /// <summary>
        /// Longitude (X), WGS84, or 0.0 if unknown
        /// </summary>
        public double Longitude { get; private set; }

        /// <summary>
        /// True if the GPS has a fix
        /// </summary>
        public bool HasFix { get; private set; }
        
        /// <summary>
        /// Satellites in view
        /// </summary>
        public int SatellitesInView { get; private set; }

        /// <summary>
        /// Heading in degrees, or 0 if none
        /// </summary>
        public float Heading { get; private set; }

        /// <summary>
        /// Height in meters over MSL, or 0 if unknown
        /// </summary>
        public float Height { get; private set; }

        /// <summary>
        /// Ground speed in km/h, or 0 if unknown
        /// </summary>
        public float GroundSpeed { get; private set; }

        /// <summary>
        /// Date and Time (UTC), or DateTime.MinValue if unknown. Might be known without fix
        /// </summary>
        public DateTime DateTime { get; private set; }

        /// <summary>
        /// Update rate in ms (default: 500ms)
        /// </summary>
        public ushort Rate
        {
            get { return _rate; }
            set { 
                _rate = value;
                SetRate(value);
            }
        }

        /// <summary>
        /// Gets/Sets the navigation model used within the GPS (default: Portable)
        /// </summary>
        public DynNavModel NavModel
        {
            get { return _navModel; }
            set { 
                _navModel = value;
                SetNavEngine(value);
            }
        }

        /// <summary>
        /// Gets/Sets the power mode used by the GPS (default: MaximumPerformance)
        /// </summary>
        public PowerSetting PowerMode
        {
            get { return _powerMode; }
            set { 
                _powerMode = value;
                SetPowerMode(_powerMode);
            }
        }

        /// <summary>
        /// NMEA Constants
        /// </summary>
        private static class NMEA
        {
            public const string RecommendedMinimumData = "RMC";
            public const string GPSFixData = "GGA";
            public const string SatellitesInView = "GSV";
        }

        /// <summary>
        /// UBX Constants
        /// </summary>
        private static class UBX
        {
            // ReSharper disable InconsistentNaming
            public const byte CFG = 0x06;
            public const byte CFG_MSG = 0x01;
            public const byte CFG_NAV5 = 0x24;
            public const byte CFG_RXM = 0x11;
            public const byte CFG_RST = 0x04;
            public const byte CFG_RATE = 0x08;
            // ReSharper restore InconsistentNaming
        }

        /// <summary>
        /// Initialises the gps module using the specified serial port
        /// </summary>
        /// <param name="serialPort">e.g. COM1</param>
        /// <param name="fullReset">True if all saved data should be erased</param>
        public CN06GPS(string serialPort, bool fullReset)
        {
            _port = new SerialPort(serialPort)
                        {
                            BaudRate = 9600,
                            Parity = Parity.None,
                            StopBits = StopBits.One,
                            DataBits = 8,
                            Handshake = Handshake.None
                        };

            _port.DataReceived += OnData;
            _port.ErrorReceived += OnError;
            _port.Open();
            Print("Init on Port "+serialPort + " -- v1.0");

            // Set up the module
            WriteNMEA("PUBX", "41", 1 /* uart1 */, Utils.BitMask(11) /* in mask=ubx,nmea */, Utils.BitMask(00) /* out mask=nmea */, 9600 /* baud */, 0); // set output mode
            Print("Resetting module ..");
            SendReset(fullReset);
            WriteNMEA("PUBX", "41", 1 /* uart1 */, Utils.BitMask(11) /* in mask=ubx,nmea */, Utils.BitMask(10) /* out mask=nmea */, 9600 /* baud */, 0); // set output mode
            SetPowerMode(_powerMode);
            SetRate(_rate);
            SetNavEngine(_navModel);
            SetNMEAMessageRate(MessageRateID.RMC, 1);
            SetNMEAMessageRate(MessageRateID.GSV, 4);
            SetNMEAMessageRate(MessageRateID.GGA, 2);
            SetNMEAMessageRate(MessageRateID.VTG, 0);
            SetNMEAMessageRate(MessageRateID.GSA, 0);
            SetNMEAMessageRate(MessageRateID.GLL, 0);
            SetNMEAMessageRate(MessageRateID.TXT, 0);
            Thread.Sleep(100);
            Print("Setup complete!");
        }

        private static void OnError(object sender, SerialErrorReceivedEventArgs e)
        {
            throw new IOException("Communication with CN-06 GPS failed ("+e.EventType+")");
        }

        private void OnData(object sender, SerialDataReceivedEventArgs e)
        {
            while (_port.BytesToRead > 0)
            {
                var c = (char)_port.ReadByte();

                if (c == '\r' || c == '\n')
                {
                    if (_read > 0)
                    {
                        OnRead(new string(_buf, 0, _read));
                        _read = 0;
                    }
                }
                else if (_read < _buf.Length && (_read > 0 || c == '$'))
                    _buf[_read++] = c;
            }
        }

        private void OnRead(string str)
        {
            if (str[0] != '$' || str.Length < 4 || str[str.Length-3] != '*')
                return; // invalid data

            Print("UART RX: " + str);
            
            // Calculate the checksum
            int checksum = (str[str.Length - 2].AsHex() << 4) + str[str.Length - 1].AsHex();
            int ourSum = 0;
            for (int i = 1; i <= str.Length-4; i++)
                ourSum ^= (byte)str[i];

            if (checksum != ourSum)
            {
                Debug.Print("CHECKSUM MISMATCH -- expected="+ourSum+" but got="+checksum+"!");
                return;
            }

            // Parse the data
            string[] parts = str.Substring(1, str.Length - 3 - 1).Split(',');
            string tag = parts[0].Substring(2);

            if (parts[0][0] != 'G' || parts[0][1] != 'P')
                return;

            switch (tag)
            {
                case NMEA.RecommendedMinimumData:
                    bool hadFix = HasFix;
                    HasFix = (Def(parts[2], "V") == "A");

                    if (HasFix)
                    {
                        if (!hadFix)
                            SetNMEAMessageRate(MessageRateID.GSV, 30);

                        double lati = double.Parse(Def(parts[3], "0.0")); // ddmm.mmmmm
                        if (parts[4] == "S")
                            lati *= -1;
                        double longi = double.Parse(Def(parts[5], "0.0")); // ddmm.mmmmm
                        if (parts[6] == "W")
                            longi *= -1;
                        Latitude = ToPercentCoords(lati);
                        Longitude = ToPercentCoords(longi);
                        GroundSpeed = (float)(double.Parse(Def(parts[7], "0.0")) * KNOTS_TO_KMH); // float, in knots
                        Heading = (float)double.Parse(Def(parts[8], "0.0")); // float, in degrees
                    }
                    else if (hadFix)
                        SetNMEAMessageRate(MessageRateID.GSV, 4);

                    if (!parts[1].IsNullOrEmpty() && !parts[9].IsNullOrEmpty()) // Parse time
                    {
                        string utcTime = Def(parts[1], "000000.00"); // hhmmss.ss
                        string date = Def(parts[9], "010100"); // ddmmyy (stops working @ year 21k!)
                        DateTime = new DateTime(2000 + ParseNum(date, 4, 2),
                                                ParseNum(date, 2, 2),
                                                ParseNum(date, 0, 2),
                                                ParseNum(utcTime, 0, 2),
                                                ParseNum(utcTime, 2, 2),
                                                ParseNum(utcTime, 4, 2),
                                                ParseNum(utcTime, 7, 2) * 10);
                    } else
                        DateTime = DateTime.MinValue;
                    break;
                case NMEA.GPSFixData:
                    Height = (float) double.Parse(Def(parts[9], "0.0"));
                    break;
                    
                case NMEA.SatellitesInView:
                    SatellitesInView = int.Parse(Def(parts[3], "0"));
                    break;
            }
        }

        #region Parse-Helpers
        private static double ToPercentCoords(double ddmm)
        {
            double dd = ddmm/100.0;
            var ddInt = (int) dd;
            return ddInt + (dd - ddInt)*100.0/60.0;
        }
        private static int ParseNum(string str, int start, int digits)
        {
            int num = 0;
            int pot = 1;
            for (int i = start + digits - 1; i >= start; i--)
            {
                num += (str[i] - '0') * pot;
                pot *= 10;
            }
            return num;
        }
        private static string Def(string str, string defValue)
        {
            return str.IsNullOrEmpty() ? defValue : str;
        }
        #endregion

        #region Writing NMEA/UBX
        /// <summary>
        /// Writes a NMEA command
        /// </summary>
        /// <param name="tag"></param>
        /// <param name="args"></param>
        private void WriteNMEA(string tag, params object[] args)
        {
            int chksum = 0;
            const byte comma = (byte) ',';
            _port.WriteByte((byte)'\r');
            _port.WriteByte((byte)'\n');
            _port.WriteByte((byte)'$');

            WriteString(tag, ref chksum);

            foreach (object o in args)
            {
                _port.WriteByte(comma); 
                chksum ^= comma;
                WriteString(o.ToString(), ref chksum);
            }

            _port.WriteByte((byte)'*');
            _port.WriteByte((byte)(chksum >> 4).ToHexLetter());
            _port.WriteByte((byte)(chksum).ToHexLetter());
            _port.WriteByte((byte)'\r');
            _port.WriteByte((byte)'\n');

            _port.Flush();
            Thread.Sleep(66);
        }

        /// <summary>
        /// Writes a UBX command
        /// </summary>
        /// <param name="clazz"></param>
        /// <param name="id"></param>
        /// <param name="buf"></param>
        private void WriteUBX(byte clazz, byte id, params byte[] buf)
        {
            // Calculate the header
            _port.WriteByte(0xFF);
            _port.WriteByte(0xB5);
            _port.WriteByte(0x62);
            _port.WriteByte(clazz);
            _port.WriteByte(id);
            _port.WriteByte((byte)(buf.Length));
            _port.WriteByte((byte)(buf.Length >> 8));
            
            // Send the actual data
            _port.Write(buf, 0, buf.Length);

            // Calculate the checksum
            byte chksumA = 0;
            byte chksumB = 0;
            FletcherCrc(ref chksumA, ref chksumB, clazz);
            FletcherCrc(ref chksumA, ref chksumB, id);
            FletcherCrc(ref chksumA, ref chksumB, (byte)buf.Length);
            FletcherCrc(ref chksumA, ref chksumB, (byte)(buf.Length >> 8));
            for (int i = 0; i < buf.Length; i++)
                FletcherCrc(ref chksumA, ref chksumB, buf[i]);
            _port.WriteByte(chksumA);
            _port.WriteByte(chksumB);

            _port.Flush();
            Thread.Sleep(66);
        }

        /// <summary>
        /// Calculates the additive Fletcher CRC
        /// </summary>
        /// <param name="chksumA"></param>
        /// <param name="chksumB"></param>
        /// <param name="data"></param>
        private static void FletcherCrc(ref byte chksumA, ref byte chksumB, byte data)
        {
            chksumA += data;
            chksumB += chksumA;
        }

        /// <summary>
        /// Writes a string into the serial port, and calculates the additive fletcher crc for that string
        /// </summary>
        /// <param name="str"></param>
        /// <param name="chksum"></param>
        private void WriteString(string str, ref int chksum)
        {
            for (int i = 0; i < str.Length; i++)
            {
                var b = (byte) str[i];
                _port.WriteByte(b);
                chksum ^= b;
            }
        }
        #endregion

        #region UBX Configuration commands
        private enum MessageRateID : ushort
        {
            // ReSharper disable InconsistentNaming
            // NMEA messages
            GGA = 0xF000,
            GLL = 0xF001,
            GSA = 0xF002,
            GSV = 0xF003,
            RMC = 0xF004,
            TXT = 0xF041,
            VTG = 0xF005,
            // ReSharper restore InconsistentNaming
        }
        
        private void SetNMEAMessageRate(MessageRateID type, byte rate)
        {
            WriteUBX(UBX.CFG, UBX.CFG_MSG, (byte)((ushort)type >> 8), (byte)type, 0, rate, 0, 0, 0, 0);
        }      
        private void SetNavEngine(DynNavModel navModel)
        {
            const int maskDynModel = (1 << 0);
            WriteUBX(UBX.CFG, UBX.CFG_NAV5, Utils.BitMask(maskDynModel), Utils.BitMask(0), (byte)navModel, 
                     0,  0,0,0,0,   0,0,0,0,  0,  0,  0,0,  0,0,  0,0,  0,0,  0, 0,   0,0,0,0,   0,0,0,0,  0,0,0,0);
        }
        private void SetRate(ushort measurementRate)
        {
            WriteUBX(UBX.CFG, UBX.CFG_RATE, (byte)measurementRate, (byte)(measurementRate >> 8), 01, 00, 0, 0);
        }
        private void SendReset(bool fullReset)
        {
            const byte swReset = 0x01;
            int mask = fullReset ? 0xFFFF : 0x0001;
            WriteUBX(UBX.CFG, UBX.CFG_RST, /* mask */ (byte)(mask), (byte)(mask >> 8), /* mode */ swReset, /* reserved */ 0x00);
            _read = 0;
            Thread.Sleep(200);
        }
        private void SetPowerMode(PowerSetting mode)
        {
            WriteUBX(UBX.CFG, UBX.CFG_RXM, 8 /* reserved */, (byte)mode);
        }
        #endregion

        private static void Print(object str)
        {
            Debug.Print("[CN-06 GPS] " + (str ?? ""));
        }

        public override string ToString()
        {
            var status = new StringBuilder("[CN-06 GPS]");
            status.Append(" InView="); status.Append(SatellitesInView);
            if (HasFix) {
                status.Append(" Lat="); status.Append(Latitude);
                status.Append(" Long="); status.Append(Longitude);
                status.Append(" Dir="); status.Append(Heading);
                status.Append(" Height="); status.Append(Height);
                status.Append(" GSpeed="); status.Append(GroundSpeed);
            } else
                status.Append(" NoFix");

            if (DateTime > DateTime.MinValue)
                status.Append(" Time="); status.Append(DateTime);
            return status.ToString();
        }
    }
}
